/*
 * score_matrix_reader.cpp
 *
 *  Created on: 2010/06/22
 *      Author: shu
 */

#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdexcept>
#include <sstream>
#include "score_matrix.h"
#include "score_matrix_reader.h"
#include "sequence.h"
#include "encoder.h"

using namespace std;

const string ScoreMatrixReader::default_matrix_name_="BLOSUM62";
const string ScoreMatrixReader::default_matrix_ = 
"   A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  B  Z  X  *\nA  4 -1 -2 -2  0 -1 -1  0 -2 -1 -1 -1 -1 -2 -1  1  0 -3 -2  0 -2 -1  0 -4\n R -1  5  0 -2 -3  1  0 -2  0 -3 -2  2 -1 -3 -2 -1 -1 -3 -2 -3 -1  0 -1 -4\n N -2  0  6  1 -3  0  0  0  1 -3 -3  0 -2 -3 -2  1  0 -4 -2 -3  3  0 -1 -4\n D -2 -2  1  6 -3  0  2 -1 -1 -3 -4 -1 -3 -3 -1  0 -1 -4 -3 -3  4  1 -1 -4\n C  0 -3 -3 -3  9 -3 -4 -3 -3 -1 -1 -3 -1 -2 -3 -1 -1 -2 -2 -1 -3 -3 -2 -4\n Q -1  1  0  0 -3  5  2 -2  0 -3 -2  1  0 -3 -1  0 -1 -2 -1 -2  0  3 -1 -4\n E -1  0  0  2 -4  2  5 -2  0 -3 -3  1 -2 -3 -1  0 -1 -3 -2 -2  1  4 -1 -4\n G  0 -2  0 -1 -3 -2 -2  6 -2 -4 -4 -2 -3 -3 -2  0 -2 -2 -3 -3 -1 -2 -1 -4\n H -2  0  1 -1 -3  0  0 -2  8 -3 -3 -1 -2 -1 -2 -1 -2 -2  2 -3  0  0 -1 -4\n I -1 -3 -3 -3 -1 -3 -3 -4 -3  4  2 -3  1  0 -3 -2 -1 -3 -1  3 -3 -3 -1 -4\n L -1 -2 -3 -4 -1 -2 -3 -4 -3  2  4 -2  2  0 -3 -2 -1 -2 -1  1 -4 -3 -1 -4\n K -1  2  0 -1 -3  1  1 -2 -1 -3 -2  5 -1 -3 -1  0 -1 -3 -2 -2  0  1 -1 -4\n M -1 -1 -2 -3 -1  0 -2 -3 -2  1  2 -1  5  0 -2 -1 -1 -1 -1  1 -3 -1 -1 -4\n F -2 -3 -3 -3 -2 -3 -3 -3 -1  0  0 -3  0  6 -4 -2 -2  1  3 -1 -3 -3 -1 -4\n P -1 -2 -2 -1 -3 -1 -1 -2 -2 -3 -3 -1 -2 -4  7 -1 -1 -4 -3 -2 -2 -1 -2 -4\n S  1 -1  1  0 -1  0  0  0 -1 -2 -2  0 -1 -2 -1  4  1 -3 -2 -2  0  0  0 -4\n T  0 -1  0 -1 -1 -1 -1 -2 -2 -1 -1 -1 -1 -2 -1  1  5 -2 -2  0 -1 -1  0 -4\n W -3 -3 -4 -4 -2 -2 -3 -2 -2 -3 -2 -3 -1  1 -4 -3 -2 11  2 -3 -4 -3 -2 -4\n Y -2 -2 -2 -3 -2 -1 -2 -3  2 -1 -1 -2 -1  3 -3 -2 -2  2  7 -1 -3 -2 -1 -4\n V  0 -3 -3 -3 -1 -2 -2 -3 -3  3  1 -2  1 -1 -2 -2  0 -3 -1  4 -3 -2 -1 -4\n B -2 -1  3  4 -3  0  1 -1  0 -3 -4  0 -3 -3 -2  0 -1 -4 -3 -3  4  1 -1 -4\n Z -1  0  0  1 -3  3  4 -2  0 -3 -3  1 -1 -3 -1  0 -1 -3 -2 -2  1  4 -1 -4\n X  0 -1 -1 -1 -2 -1 -1 -1 -1 -1 -1 -1 -1 -1 -2  0  0 -2 -1 -1 -1 -1 -1 -4\n * -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4  1\n";
ScoreMatrixReader::ScoreMatrixReader(const Encoder &encoder) :
  encoder_(encoder) {
}

ScoreMatrixReader::~ScoreMatrixReader() {
}

ScoreMatrix ScoreMatrixReader::Read(const string &filename) {
  ifstream in;
  in.open(filename.c_str());
  if (!in) {
    throw invalid_argument(string("can't open ") + filename + string("."));
  }

  string name = GetName(filename);
  int number_letters = GetNumberLetters();
  int matrix_size = GetMatrixSize();
  const int max_matrix_size = 4096;
  int matrix[max_matrix_size];
  assert(matrix_size <= max_matrix_size);
  Parse(in, matrix);
  return ScoreMatrix(name, matrix, number_letters);
}


ScoreMatrix ScoreMatrixReader::ReadDefault() {
  istringstream in(default_matrix_);
  string name = default_matrix_name_;
  int number_letters = GetNumberLetters();
  int matrix_size = GetMatrixSize();
  const int max_matrix_size = 4096;
  int matrix[max_matrix_size];
  assert(matrix_size <= max_matrix_size);
  Parse(in, matrix);
  return ScoreMatrix(name, matrix, number_letters);
}

list<string> ScoreMatrixReader::Split(string str, string delim) {
  list<string> result;
  string::size_type index;
  index = str.find_first_of(delim);
  while (index != string::npos) {
    if (index > 0) {
      result.push_back(str.substr(0, index));
    }
    str = str.substr(index + 1);
    index = str.find_first_of(delim);
  }
  if (str.length() > 0) {
    result.push_back(str);
  }
  return result;
}

int ScoreMatrixReader::GetNumberLetters() {
  return  encoder_.GetMaxCode() + 1;
}

int ScoreMatrixReader::GetMatrixSize() {
  int number_letters = GetNumberLetters();
  return number_letters * number_letters;
}

void ScoreMatrixReader::Parse(istream &in, int *matrix) {
  int number_letters = GetNumberLetters();
  int matrix_size = GetMatrixSize();
  string line;
  int line_number = 0;
  const unsigned int max_heading_size = 1024;
  char column_headings[max_heading_size];
  char row_headings[max_heading_size];
  for (int i = 0; i < matrix_size; ++i) {
    matrix[i] = 0;
  }

  while (!in.eof()) {
    std::getline(in, line);
    if (line[0] != '\0' && line[0] != '#') {
      //assert(line_number <= max_heading_size + 1);
      list<string> str_list = Split(line, " ");
      assert(str_list.size() <= max_heading_size);
      int i = 0;
      for (list<string>::iterator it = str_list.begin(); it != str_list.end(); ++it, ++i) {
        if (line_number == 0) {
          column_headings[i] = it->c_str()[0];
        } else if (i == 0) {
          row_headings[line_number - 1] = it->c_str()[0];
        } else {
          int value = atoi(it->c_str());
          if (!encoder_.IsUnknown(row_headings[line_number - 1])
              && !encoder_.IsUnknown(column_headings[i - 1])) {
            matrix[encoder_.Encode(row_headings[line_number - 1])
                * number_letters + encoder_.Encode(column_headings[i - 1])]
                = value;
          }
        }
      }
      ++line_number;
    }
  }
}

string ScoreMatrixReader::GetName(const string &filename) {
  string name = filename;
  string::size_type index = name.find_last_of('/');
  if (index != string::npos) {
    name = name.substr(index + 1, name.length() - index - 1);
  }
  return name;
}
